import os
import time
from functools import partial
from multiprocessing import Pool
from subprocess import Popen
import healpy as hp
import numpy as np
from astropy import table
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS

from .. import PACKAGE_PATH
from ..core.processor import CsstProcessor
from csst.msc.data_manager import CsstMscDataManager

CONFIG_PATH = PACKAGE_PATH + "/msc/pos_calib_config/"


class CsstProcMscPositionCalibration(CsstProcessor):

    def join_data(self, img_list, wht_list, flg_list):
        """
        Prepare data for running scamp; Combine all image data, weight files, flag files to their one frame.

        Parameters
        ----------
        img_list:
            image files to join together, e.g.,MSC_210304093000_0000000_06_img.fits
        wht_list:
            weith files to join together, e.g.,MSC_210304093000_0000000_06_img.fits
        flg_list:
            flag files to join together, e.g.,e.g.,MSC_210304093000_0000000_06_flg.fits
        Returns
        -------
        The joined multi-extension file(not stacked), weight, flag files.
        e.g., combined_img.fits,combined_wht.fits, combined_flg.fits.

        """

        output_imgnm = self.dm.pc_combined_file("img","fits")
        hdul_img = fits.HDUList()
        for i in range(0, len(img_list)):
            h0 = fits.PrimaryHDU(header=img_list[i][0].header)
            h1 = fits.ImageHDU(data=img_list[i][1].data, header=img_list[i][1].header)
            hdul_img.append(h0)
            hdul_img.append(h1)
        hdul_img.writeto(output_imgnm, overwrite=True)

        output_whtnm = self.dm.pc_combined_file("wht","fits")
        hdul_wht = fits.HDUList()
        for i in range(0, len(wht_list)):
            h0 = fits.PrimaryHDU(header=wht_list[i][0].header)
            h1 = fits.ImageHDU(data=wht_list[i][1].data, header=wht_list[i][1].header)
            hdul_wht.append(h0)
            hdul_wht.append(h1)
        hdul_wht.writeto(output_whtnm, overwrite=True)

        output_flgnm = self.dm.pc_combined_file("flg","fits")
        hdul_flg = fits.HDUList()
        for i in range(0, len(flg_list)):
            h0 = fits.PrimaryHDU(header=flg_list[i][0].header)
            h1 = fits.ImageHDU(data=flg_list[i][1].data, header=flg_list[i][1].header)
            hdul_flg.append(h0)
            hdul_flg.append(h1)
        hdul_flg.writeto(output_flgnm, overwrite=True)

    def run_sextractor(self, fn_list, ccd_ids):
        """
        Run sextractor

        Parameters 
        ----------
        ccd_ids:
            ccd list...
        Returns
        -------
        The photometric catalog, with position and flux, e.g.,MSC_210304093000_0000000_06_img.cat
        """
        cat_list = []
        fn_list = []
        for this_ccd_id in ccd_ids:
            print(type(this_ccd_id),this_ccd_id)
            this_ccd_id=float(this_ccd_id)
            this_ccd_id=int(this_ccd_id)
            sex_cat = self.dm.l1_sci(ccd_id = this_ccd_id, suffix="img", ext="cat")
            fn = self.dm.l1_sci(ccd_id=this_ccd_id, suffix="img", ext="fits")
            cat_list.append(sex_cat)
        
        config_sextractor = CONFIG_PATH + "new_csst_realtime.no.weight.sex"
        sex_comd1 = 'sex -c ' + config_sextractor + ' '
        sex_comd2 = fn_list + ' -CATALOG_NAME ' + cat_list
        sex_comd3 = ' -PARAMETERS_NAME ' + CONFIG_PATH + 'csst_realtime.param' + ' -FILTER_NAME ' + CONFIG_PATH + 'csst_realtime.conv' + ' -STARNNW_NAME ' + CONFIG_PATH + 'csst_realtime.nnw'
        sex_comd = sex_comd1 + sex_comd2 + sex_comd3
        print(sex_comd)
        p = Popen(sex_comd, shell=True)
        p.wait()

    def combine_catalog(self, ccd_ids):
        """
        Combine the sextractor catalog together

        Parameters
        -----------
        ccd_ids: ccd list
        Returns
        -------
        The combined catalog,e.g., combined_cat.fits
        """
        fn = self.dm.pc_combined_file("cat","fits")
        hdul = fits.HDUList()
        for this_ccd_id in ccd_ids:
            cat_nm = self.dm.l1_sci(ccd_id=this_ccd_id, suffix="img", ext="cat")
            cat_i = fits.open(cat_nm)
            data=cat_i[2].data
            msk=(data['MAGERR_AUTO']<0.2)&(data['CLASS_STAR']>0.5)&(data['CLASS_STAR']<=1.0)
            data=data[msk]
            cat_i[2].data=data
            hdul.append(cat_i[0])
            hdul.append(cat_i[1])
            hdul.append(cat_i[2])
        hdul.writeto(fn, overwrite=True)
 
    def run_scamp(self, wcs_refine=False):
        """
        Run scamp
        Returns
        -------
        Image header updated with WCS keywords, combined_cat.head.
        """
        cb_fn = self.dm.pc_combined_file("cat","fits")
        config_scamp = CONFIG_PATH + "csst.scamp"
        scamp_comd = 'scamp ' + cb_fn + ' -ASTREFCAT_NAME= ' + 'ref.cat\
        -MERGEDOUTCAT_NAME ' + 'merged.cat -FULLOUTCAT_NAME ' + 'full.cat\
        -c ' + config_scamp
        print(scamp_comd)
        p = Popen(scamp_comd, shell=True)
        p.wait()
        if wcs_refine:
            Popen('cp '+ self.dm.pc_combined_file("cat","head")+" "+self.dm.pc_combined_file("cat","ahead"),shell=True)
            scamp_comd = 'scamp ' + cb_fn + ' -ASTREFCAT_NAME= ' + 'ref.cat\
            -MERGEDOUTCAT_NAME ' + 'merged.cat -FULLOUTCAT_NAME ' + 'full.cat\
            -c ' + config_scamp + '-AHEADER_SUFFIX '+ self.dm.pc_combined_file("cat","ahead")
            print(scamp_comd)
            p=Popen(scamp_comd,shell=True)
            p.wait()


    def convert_hdu_to_ldac(self, hdu):
        """
        Convert an hdu table to a fits_ldac table (format used by astromatic suite)

        Parameters
        ----------
        hdu:
            `astropy.io.fits.BinTableHDU` or `astropy.io.fits.TableHDU`
            HDUList to convert to fits_ldac HDUList

        Returns
        -------
        tbl1:
            `astropy.io.fits.BinTableHDU`
            Header info for fits table (LDAC_IMHEAD)
        tbl2:
            `astropy.io.fits.BinTableHDU`
            Data table (LDAC_OBJECTS)
        """
        tblhdr = np.array([hdu[1].header.tostring()])
        col1 = fits.Column(name='Field Header Card', array=tblhdr, format='13200A')
        cols = fits.ColDefs([col1])
        tbl1 = fits.BinTableHDU.from_columns(cols)
        tbl1.header['TDIM1'] = '(80, {0})'.format(len(hdu[1].header))
        tbl1.header['EXTNAME'] = 'LDAC_IMHEAD'

        dcol = fits.ColDefs(hdu[1].data)
        tbl2 = fits.BinTableHDU.from_columns(dcol)
        tbl2.header['EXTNAME'] = 'LDAC_OBJECTS'
        return tbl1, tbl2

    def get_refcat(self, img_list, path_gaia, search_radius, silent=True,pm_correct=True):
        """
        Get reference catalog for scamp. The reference cat is GAIA EDR3.

        Parameters
        ----------
        image_prefix:
            a image to get its reference catalog, e.g.,combined_img.fits.
            Usually the center of the image is the wcs parameters CRVAL1,CRVAL1.
        search_radius:
            circle radius for searching, units: degree. e.g., 2 degree for a 1x1 deg^2 image.
            For large ccd size, use larger radius. csst, r=3 deg.
        path_gaia: directory of the reference catalog.

        Returns
        -------
        outcat:
            filename of the cross matched catalog: ref.cat
        """
        outcat = self.dm.pc_ref_cat
        fnmae =  cb_fn = self.dm.pc_combined_file("img","fits")
        hdu = fits.open(fname)
        header1 = hdu[0].header
        header2 = hdu[1].header
        if (pm_correct):
            oday=header1['DATE-OBS']
            otime=header1['TIME-OBS']
            exptime=header1['EXPTIME']
            odaytime=header1['DATE-OBS']+'T'+header1['TIME-OBS']
            t = Time(oday, format='isot', scale='utc')
            t.format = 'decimalyear'
            t1=Time(2016.0, format='decimalyear', scale='utc')
            deltatime= t.value -2016.0
            print('(Time - 2016.0) = deltatime =', deltatime)
        else:   
            deltatime=0.00  
        ra = float(header2['CRVAL1'])
        dec = float(header2['CRVAL2'])
        c = SkyCoord(ra, dec, unit=(u.deg, u.deg))
        print('ra, dec ra.deg dec.deg= ', ra, dec, c.ra.deg, c.dec.deg)
        ra = c.ra.deg
        dec = c.dec.deg
        c = SkyCoord(ra, dec, unit=(u.deg, u.deg))
        phi = c.ra.deg / (180. / np.pi)
        theta = (90. - c.dec.deg) / (180 / np.pi)
        vec = hp.ang2vec(theta, phi)
        list1 = hp.query_disc(nside=32, vec=vec, radius=np.radians(search_radius))
        if -1 in list1:
            list1.remove(-1)
        ipring = np.array(list1)
        pix = np.unique(ipring)
        npix = pix.size
        print(ipring, 'ipring', pix, 'pix', npix, 'npix')
        dt = np.dtype(
            [('ra', '>f8'), ('dec', '>f8'), ('ra_error', '>f8'), ('dec_error', '>f8'), ('phot_g_mean_mag', '>f8'),
             ('pmra', '>f8'), ('pmra_error', '>f8'), ('pmdec', '>f8'), ('pmdec_error', '>f8')])
        refcat = table.Table(dtype=dt)
        for i in pix:
            print('i= %5.5d' % i, path_gaia)
            fname = path_gaia + 'healpix-' + '%5.5d' % i + '.fits'
            print('fname=', fname)
            if not silent: print('Reading ', fname)
            d = table.Table.read(fname)
            refcat = [refcat, d]
            refcat = table.vstack(refcat, join_type='inner')
        refcat.rename_column('ra', 'X_WORLD')
        refcat.rename_column('dec', 'Y_WORLD')

        mask = (refcat['pmdec'] != refcat['pmdec'])
        refcat['pmdec'][mask] = 0
        mask = (refcat['pmra'] != refcat['pmra'])
        refcat['pmra'][mask] = 0
        refcat['X_WORLD'] = refcat['X_WORLD'] + deltatime * refcat['pmra'] / np.cos(
            refcat['Y_WORLD'] / 180. * np.pi) / 3600.0 / 1000.0
        refcat['Y_WORLD'] = refcat['Y_WORLD'] + deltatime * refcat['pmdec'] / 3600.0 / 1000.0
        refcat['ra_error'] = refcat['ra_error'] / 1000.0 / 3600.0
        refcat['dec_error'] = refcat['dec_error'] / 1000.0 / 3600.0
        refcat.rename_column('ra_error', 'ERRA_WORLD')
        refcat.rename_column('dec_error', 'ERRB_WORLD')
        refcat.rename_column('phot_g_mean_mag', 'MAG')

        hdu = refcat
        hdu1 = self.convert_hdu_to_ldac(hdu)
        hdup = fits.PrimaryHDU()
        hdu = hdu1[0]
        tbhdu = hdu1[1]
        thdulist = fits.HDUList([hdup, hdu, tbhdu])
        thdulist.writeto(outcat)

        print('##################### end #####################')

    def rewrite_wcs_head(self, head):
        """
        Rewrite the WCS head from Scamp to the standard fits header

        Parameters
        ----------
        head: scamp head file names

        Returns
        -------
        wcshead: a new head file in fits format

        """
        wcshead = head + '.fits'
        f = open(head, 'r')
        f1 = open(wcshead, 'w')
        a = ''
        i = 0
        for v in f.readlines():
            sp = ''
            asp = ''
            i += 1
            if len(v) <= 81:
                sp = ' ' * (81 - len(v))
            if 'END' in v:
                asp = ' ' * 80 * (36 - i % 36)
                i = i + (36 - i % 36)
                # print(i)
            a = a + v + sp + asp
        f1.write(a.replace('\n', ''))
        f1.close()
        f.close()
        return wcshead

    def check_astrometry(self, img_list):
        """
        Check position calibration quality

        Parameters
        ------------
        img_list:
            list of images, in table format

        Returns
        -------
        ccdraoff:
            ra difference between observed catalog and reference catalog
        ccddecoff:
            dec difference between observed catalog and reference catalog
        ccdraoff_med,ccddecoff_med:
            median of the ra or dec difference
        ccdraoff_rms,ccddecoff_rms:
            rms of the ra or dec difference
        """
        print('############## check the astrometry quality and save files ################')
        r1 = []
        d1 = []
        fn_head = self.dm.pc_combined_file("cat","head")
        fn_scat = self.dm.pc_combined_file("cat","fits")
        data_scat = fits.open(fn_scat)

        wcshead = self.rewrite_wcs_head(fn_head)
        scat = fits.open(fn_scat)
        hdul = fits.HDUList()
        for i in range(0, len(img_list)):
            wcshdr = fits.getheader(wcshead, i, ignore_missing_simple=True)       
            wcshdr['CTYPE1'] = 'RA---TPV'
            wcshdr['CTYPE2'] = 'DEC--TPV'
            w = WCS(wcshdr)
            sexcat = data_scat[i*3+2].data
            x = sexcat['XWIN_IMAGE']
            y = sexcat['YWIN_IMAGE']
            r, d = w.all_pix2world(x, y, 0)  
            sexcat['ALPHA_J2000'] = r
            sexcat['DELTA_J2000'] = d
            r1 = np.hstack((r1, r))
            d1 = np.hstack((d1, d))
        obsc = SkyCoord(ra=r1 * u.degree, dec=d1 * u.degree)
        tmp_cat = np.zeros((len(obsc), 2))
        tmp_cat[:, 0] = obsc.ra
        tmp_cat[:, 1] = obsc.dec
        np.savetxt(dm.pc_scamp_coord, tmp_cat, fmt="%.10f %.10f", delimiter="\n")

        gaia_cat=Table.read(self.dm.pc_ref_cat) 
        gaia_ra=gaia_cat['X_WORLD']
        gaia_dec=gaia_cat['Y_WORLD']
        refc=SkyCoord(ra=gaia_ra*u.degree,dec=gaia_dec*u.degree)
        idx, d2d, d3d = obsc.match_to_catalog_sky(refc)
        ref_uid=np.unique(idx)
        obs_uid=np.full_like(ref_uid,-1)
        tmpj=-1
        ccdraoff_med=ccddecoff_med=ccdra_rms=ccddec_rms=-1
        for i in ref_uid:
            tmpj=tmpj+1
            iid=(idx == i)
            iiid = (d2d.deg[iid] == d2d.deg[iid].min())
            obs_uid[tmpj]=iid.nonzero()[0][iiid.nonzero()[0]][0]

        uidlim=d2d[obs_uid].arcsecond <1. # set match radius=1 arcsec
        if uidlim.sum()>0:
            obs_uidlim=obs_uid[uidlim]
            ref_uidlim=ref_uid[uidlim]
            ccdraoff=(obsc[obs_uidlim].ra- refc[ref_uidlim].ra).arcsec*np.cos(obsc[obs_uidlim].dec.deg*np.pi/180.)
            ccdraoff_med=np.median(ccdraoff)
            ccdra_rms=np.std(ccdraoff)
            ccddecoff=(obsc[obs_uidlim].dec- refc[ref_uidlim].dec).arcsec
            ccddec_rms=np.std(ccddecoff)
            ccddecoff_med=np.median(ccddecoff)
            match_num=len(ccdraoff)
            print('################# astrometry result: ##############')
            if (match_num<100):
                print('### bad astrometry ###')
        print('median ra_off, dec_off (mas) from scamp:',ccdraoff_med*1000.,ccddecoff_med*1000.)  
        print('rms ra_off, dec_off (mas) from scamp:',ccdra_rms*1000.,ccddec_rms*1000.)  
        print('############################################')    
        return ccdraoff,ccddecoff,ccdraoff_med,ccddecoff_med,ccdra_rms,ccddec_rms

    def write_headers(self, ccd_ids):
        """
        Wrtie history to header
        """
        head_suffix = self.dm.pc_combined_head_fits
        hdul2 = fits.open(head_suffix, ignore_missing_simple=True)
        for this_ccd_id in ccd_ids:
             fits_nm = self.dm.l1_sci(this_ccd_id, suffix="img", ext="head")
             hdul1 = fits.open(fits_nm, mode='update', ignore_missing_simple=True)
             hdr = hdul1[0].header
             hdr2 = hdul2[i].header
             hdr.extend(hdr2, unique=True, update=True)
             WCS_S = 0
             WCS_V = '2.0.4'
             WCS_P = 'default.scamp'
             WCS_TOL = time.strftime('%Y-%m-%d %H:%M:%S %p')
             hdr.set('WCS_S', '0', '0=done')
             hdr.set('WCS_V', WCS_V, 'Version of WCS calibration')
             hdr.set('WCS_P', WCS_P, 'Configure file name of WCS')
             hdr.set('WCS_TOL', WCS_TOL, 'Time of last wcs calibration')
             hdul1.flush()
             hdul1.close()

    def make_plots(self, ccdoff):
        print('##### Analyzing the scampe result, making some pltos.... ####')
        plt.figure(figsize=(11,5))
        ax1=plt.subplot(121)
        bin=0.05
        plt.grid(color='grey',ls='--')
        plt.plot(ccdoff[0],ccdoff[1],'ko',markersize=3,alpha=0.3)
        plt.xlabel(r'$\Delta$ RA (arcsec)',fontsize=12)
        plt.ylabel(r'$\Delta$ Dec (arcsec)',fontsize=12)
        ax2=plt.subplot(122)
        plt.grid(color='grey',ls='--')
        plt.hist(ccdoff[0],bins=np.arange(-1,1, bin),histtype="step",color="r",label=r'$\Delta$RA (arcsec)')
        plt.hist(ccdoff[1],bins=np.arange(-1,1, bin),histtype="step",color="b",label=r'$\Delta$Dec (arcsec)')
        plt.legend()
        a=str(float(ccdoff[2]))
        b=str(float(ccdoff[4]))
        c=str(float(ccdoff[3]))
        d=str(float(ccdoff[5]))
        plt.text(-0.95,45,r'$\mu$='+a[0:6]+r',  $\sigma$='+b[0:5]+' (arcsec)',color='red')
        plt.text(-0.95,35,r'$\mu$='+c[0:6]+r',  $\sigma$='+d[0:5]+' (arcsec)',color='blue')
        plt.xlabel('coord_diff (arcsec)',fontsize=12)            
        plt.savefig(self.dm.pc_radecoff,dpi=300)


    def prepare(self, dm):
        self.dm = dm

    def run(self, img_list, wht_list, flg_list, fn_list, ccd_ids, path_gaia, search_radius, pm_correct=True, wcs_refine=False, plot=False):
        print('preparing files for position calibration....')
        #self.join_data(img_list, wht_list, flg_list)
        print('################## run sextractor ###################')
        p = Pool()
        prod_x = partial(self.run_sextractor,ccd_ids)
        result = p.map(prod_x, fn_list)
        p.close()
        p.join()
        self.run_sextractor(ccd_ids)
        print('################## sextractor done ###################')
        print('############### combine sextractor catalog ###############')
        self.combine_catalog(ccd_ids)
        print('############### get reference catalog ###############3')
        self.get_refcat(img_list, path_gaia=path_gaia, search_radius=search_radius, silent=True,pm_correct=True)
        print('############### run scamp ##################')
        self.run_scamp(wcs_refine=wcs_refine)
        print('################ scamp done #################')
        print('Checking astrometry quality....')
        ccdoff = self.check_astrometry(img_list)
        print('################ updating headers.... #############')
        self.write_headers(ccd_ids)
        print('#### Position calibration process done ####')

        if plot:
            make_plot(ccdoff)
                    
    def cleanup(self):
        pass

